#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <sys/select.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/glx.h>
#include "ex.h"

#ifdef USE_XRANDR
#include <X11/extensions/Xrandr.h>

char *randr_err_msg[] = {
	"RRSetConfigSuccess",
	"RRSetConfigInvalidConfigTime",
	"RRSetConfigInvalidTime",
	"RRSetConfigFailed"
};
#endif

#ifdef USE_XF86VM
#include <X11/extensions/xf86vmode.h>
#endif

static int video_mode_switch(int xsz, int ysz);
static int get_size(Display *dpy, int scr, int *xsz, int *ysz);
static int set_size(Display *dpy, int scr, int xsz, int ysz);

static Display *dpy;
static int scr;
static Window win;
static GLXContext ctx;
static unsigned int event_mask;

static Atom wm_prot, wm_del_win;

/* window stuff */
static int win_is_fullscr;
static int win_xsz, win_ysz;
static int orig_xsz, orig_ysz;	/* save original resolution if we switch */

/* event callbacks */
void (*disp_handler)(void);
void (*reshape_handler)(int, int);
void (*keyb_handler)(unsigned int, int);
void (*mouse_handler)(int, int, int, int);
void (*motion_handler)(int, int);

int create_gfx(int xsz, int ysz, int fullscreen)
{
	Window root;
	XVisualInfo *vis;
	XSetWindowAttributes xattr;
	XClassHint class_hint;
	unsigned int attr_valid_mask = CWColormap | CWBackPixel | CWBorderPixel;

	int glattr[] = {
		GLX_RGBA, GLX_DOUBLEBUFFER,
		GLX_RED_SIZE, 8,
		GLX_GREEN_SIZE, 8,
		GLX_BLUE_SIZE, 8,
		GLX_DEPTH_SIZE, 24,
		GLX_STENCIL_SIZE, 8,
		None
	};

	win_xsz = xsz;
	win_ysz = ysz;
	win_is_fullscr = fullscreen;

	/* open connection to the X server */
	if(!(dpy = XOpenDisplay(0))) {
		fprintf(stderr, "Failed to connect to the X server");
		return -1;
	}

	wm_prot = XInternAtom(dpy, "WM_PROTOCOLS", False);
	wm_del_win = XInternAtom(dpy, "WM_DELETE_WINDOW", False);

	scr = DefaultScreen(dpy);
	root = RootWindow(dpy, scr);

	if(!(vis = glXChooseVisual(dpy, scr, glattr))) {
		fprintf(stderr, "failed to find a suitable visual\n");
		return -1;
	}

	if(!(ctx = glXCreateContext(dpy, vis, 0, True))) {
		fprintf(stderr, "failed to create GLX context\n");
		XFree(vis);
		return -1;
	}

	xattr.background_pixel = xattr.border_pixel = BlackPixel(dpy, scr);
	xattr.colormap = XCreateColormap(dpy, root, vis->visual, AllocNone);

	if(fullscreen) {
		if(video_mode_switch(xsz, ysz) == -1) {
			/* if we failed to change the screen size, make a big window instead */
			XWindowAttributes rattr;
			XGetWindowAttributes(dpy, root, &rattr);

			xsz = rattr.width;
			ysz = rattr.height;
		}

		/* for fullscreen windows, also set override-redirect */
		xattr.override_redirect = True;
		attr_valid_mask |= CWOverrideRedirect;
	}

	if(!(win = XCreateWindow(dpy, root, 0, 0, xsz, ysz, 0, vis->depth, InputOutput,
					vis->visual, attr_valid_mask, &xattr))) {
		fprintf(stderr, "Failed to create window\n");
		XFree(vis);
		return -1;
	}
	XFree(vis);

	/* set a minimal window event mask */
	event_mask = StructureNotifyMask;
	XSelectInput(dpy, win, event_mask);

	/* let the window manager know that we are willing to participate to the
	 * WM_DELETE_WINDOW protocol.
	 */
	XSetWMProtocols(dpy, win, &wm_del_win, 1);

	set_window_title("mindlapse / X11");
	
	/* set the WM_CLASS property */
	class_hint.res_name = "mlapse";	/* TODO: get this properly */
	class_hint.res_class = "mlapse";
	XSetClassHint(dpy, win, &class_hint);

	if(glXMakeCurrent(dpy, win, ctx) == False) {
		fprintf(stderr, "glXMakeCurrent failed\n");
		glXDestroyContext(dpy, ctx);
		XDestroyWindow(dpy, win);
		return -1;
	}

	XMapWindow(dpy, win);
	return 0;
}

void destroy_gfx(void)
{
	glXDestroyContext(dpy, ctx);
	XDestroyWindow(dpy, win);
	glXMakeCurrent(dpy, None, 0);

	/* restore original screen size if appropriate */
	if(orig_xsz && orig_ysz) {
		if(set_size(dpy, scr, orig_xsz, orig_ysz) == -1) {
			fprintf(stderr, "failed to restore original screen size (%dx%d)\n", orig_ysz, orig_ysz);
		}
	}
}

void swap_buffers(void)
{
	glXSwapBuffers(dpy, win);
}

void set_window_title(const char *title)
{
	XTextProperty wm_name;
	XStringListToTextProperty((char**)&title, 1, &wm_name);
	XSetWMName(dpy, win, &wm_name);		/* set WM_NAME property */
	XSetWMIconName(dpy, win, &wm_name);	/* set WM_ICON_NAME property */
	XFree(wm_name.value);
}

void wait_event(void)
{
	fd_set rdset;
	int xfd = ConnectionNumber(dpy);

	XFlush(dpy);

	FD_ZERO(&rdset);
	FD_SET(xfd, &rdset);

	/* select may return prematurely if there is a signal.
	 * that's intentional.
	 */
	select(xfd + 1, &rdset, 0, 0, 0);
}

int process_events(void)
{
	static int win_mapped;
	KeySym sym;

	while(XPending(dpy)) {
		XEvent event;
		XNextEvent(dpy, &event);

		switch(event.type) {
			/* window becomes visible (XMapWindow) */
		case MapNotify:
			win_mapped = 1;
			if(win_is_fullscr) {	/* XXX I don't think this is necessary */
				XSetInputFocus(dpy, win, RevertToParent, CurrentTime);
			}
			break;

			/* window becomes hidden (XUnmapWindow) */
		case UnmapNotify:
			win_mapped = 0;
			break;

			/* window moved or resized, or whatever (stacking order change, etc) */
		case ConfigureNotify:
			if(event.xconfigure.width != win_xsz || event.xconfigure.height != win_ysz) {
				win_xsz = event.xconfigure.width;
				win_ysz = event.xconfigure.height;
				reshape_handler(win_xsz, win_ysz);
			}
			break;

			/* expose events arrive when part of the window needs redrawing */
		case Expose:
			/* only redraw if the window is visible and this is the final expose event */
			if(win_mapped && event.xexpose.count == 0) {
				disp_handler();
			}
			break;

			/* ICCCM WM_DELETE_WINDOW protocol event comes as a ClientMessage */
		case ClientMessage:
			if(event.xclient.message_type == wm_prot) {
				if(event.xclient.data.l[0] == wm_del_win) {
					return 0;	/* delete window was requested, return 0 to let the mainloop exit */
				}
			}
			break;

			/* keyboard events */
		case KeyPress:
		case KeyRelease:
			sym = XLookupKeysym((XKeyEvent*)&event.xkey, 0);
			keyb_handler(sym & 0xff, event.type == KeyPress ? 1 : 0);	/* TODO: add modifiers */
			break;

			/* mouse button events */
		case ButtonPress:
		case ButtonRelease:
			mouse_handler(event.xbutton.button, event.type == ButtonPress ? 1 : 0,
					event.xbutton.x, event.xbutton.y);
			break;

			/* mouse moved */
		case MotionNotify:
			motion_handler(event.xmotion.x, event.xmotion.y);
			break;

		default:
			break;
		}
	}
	return 1;
}

void post_redisplay(void)
{
	XEvent ev;

	ev.type = Expose;
	ev.xexpose.x = ev.xexpose.y = 0;
	ev.xexpose.width = win_xsz;
	ev.xexpose.height = win_ysz;
	ev.xexpose.count = 0;

	XSendEvent(dpy, win, False, 0, &ev);
}

void reshape_func(void (*reshape)(int, int))
{
	reshape_handler = reshape;

	/* notify the X server that we're interested in move/resize events */
	event_mask |= ConfigureNotify;
	XSelectInput(dpy, win, event_mask);
}

void display_func(void (*disp)(void))
{
	disp_handler = disp;

	/* notify the X server that we're interested in exposure (redraw) events */
	event_mask |= ExposureMask;
	XSelectInput(dpy, win, event_mask);
}

void keyboard_func(void (*keyb)(unsigned int, int))
{
	keyb_handler = keyb;

	/* notify the X server that we're interested in keyboard events */
	event_mask |= KeyPressMask | KeyReleaseMask;
	XSelectInput(dpy, win, event_mask);
}

void mouse_func(void (*mouse)(int, int, int, int))
{
	mouse_handler = mouse;

	/* notify the X server that we're interested in mouse button events */
	event_mask |= ButtonPressMask | ButtonReleaseMask;
	XSelectInput(dpy, win, event_mask);
}

void motion_func(void (*motion)(int, int))
{
	motion_handler = motion;

	/* notify the X server that we're interested in mouse motion events */
	event_mask |= PointerMotionMask;
	XSelectInput(dpy, win, event_mask);
}


static int video_mode_switch(int xsz, int ysz)
{
	get_size(dpy, scr, &orig_xsz, &orig_ysz);

	if(set_size(dpy, scr, xsz, ysz) == -1) {
		return -1;
	}

	printf("resized screen: %dx%d -> %dx%d\n", orig_xsz, orig_ysz, xsz, ysz);
	return 0;
}

static int get_size(Display *dpy, int scr, int *xsz, int *ysz)
{
	Window root;
	int event_base, error_base;

	scr = DefaultScreen(dpy);
	root = RootWindow(dpy, scr);

	/* first try r&r */
#ifdef USE_XRANDR
	if(XRRQueryExtension(dpy, &event_base, &error_base)) {
		XRRScreenConfiguration *xrr_config;
		XRRScreenSize *ssizes;
		Rotation rot;
		int ssize_count, curr;

		if(!(xrr_config = XRRGetScreenInfo(dpy, root))) {
			fprintf(stderr, "failed to retrieve screen configuration\n");
			return -1;
		}
		ssizes = XRRConfigSizes(xrr_config, &ssize_count);
		curr = XRRConfigCurrentConfiguration(xrr_config, &rot);

		*xsz = ssizes[curr].width;
		*ysz = ssizes[curr].height;

		XRRFreeScreenConfigInfo(xrr_config);
		return 0;
	}
#endif

	/* then try xf86vm */
#ifdef USE_XF86VM
	if(XF86VidModeQueryExtension(dpy, &event_base, &error_base)) {
		XF86VidModeModeLine modeline;
		int dot_clock;

		if(!XF86VidModeGetModeLine(dpy, scr, &dot_clock, &modeline)) {
			fprintf(stderr, "failed to get current video mode\n");
			return -1;
		}

		*xsz = modeline.hdisplay;
		*ysz = modeline.vdisplay;

		if(modeline.privsize) {
			XFree(modeline.private);
		}
		return 0;
	}
#endif

	fprintf(stderr, "can't get any size information on this X server\n");
	return -1;
}


static int set_size(Display *dpy, int scr, int xsz, int ysz)
{
	int i, event_base, error_base;
	Window root;

	root = RootWindow(dpy, scr);

	/* first try r&r */
#ifdef USE_XRANDR
	if(XRRQueryExtension(dpy, &event_base, &error_base)) {
		Status st;
		XRRScreenConfiguration *xrr_config;
		XRRScreenSize *ssizes;
		Rotation rot;
		int ssize_count, curr;
		Time timestamp, cfg_timestamp;

		if(!(xrr_config = XRRGetScreenInfo(dpy, root))) {
			fprintf(stderr, "failed to retrieve screen configuration\n");
			return -1;
		}
		ssizes = XRRConfigSizes(xrr_config, &ssize_count);
		curr = XRRConfigCurrentConfiguration(xrr_config, &rot);
		timestamp = XRRConfigTimes(xrr_config, &cfg_timestamp);

		if(xsz == ssizes[curr].width && ysz == ssizes[curr].height) {
			XRRFreeScreenConfigInfo(xrr_config);
			return 0;
		}

		for(i=0; i<ssize_count; i++) {
			if(ssizes[i].width == xsz && ssizes[i].height == ysz) {
				break;
			}
		}
		if(i == ssize_count) {
			fprintf(stderr, "requested size unsuppored by the X server configuration\n");
			XRRFreeScreenConfigInfo(xrr_config);
			return -1;
		}

		if((st = XRRSetScreenConfig(dpy, xrr_config, root, i, rot, timestamp)) != 0) {
			fprintf(stderr, "failed to set resolution with r-and-r, error: %d (%s)\n", st, randr_err_msg[st]);
			XRRFreeScreenConfigInfo(xrr_config);
#ifdef USE_XF86VM
			printf("retrying with xf86vm...\n");
			goto retry_with_xf86vm;
#endif
			return -1;
		}

		XRRFreeScreenConfigInfo(xrr_config);
		return 0;
	}
#endif

	/*try xf86vm */
#ifdef USE_XF86VM
retry_with_xf86vm:
	if(XF86VidModeQueryExtension(dpy, &event_base, &error_base)) {
		XF86VidModeModeInfo **modes;
		int mode_count;

		if(!XF86VidModeGetAllModeLines(dpy, scr, &mode_count, &modes)) {
			fprintf(stderr, "failed to get list of available video modes\n");
			return -1;
		}

		for(i=0; i<mode_count; i++) {
			if(modes[i]->hdisplay == xsz && modes[i]->vdisplay == ysz) {
				break;
			}
		}
		if(i == mode_count) {
			fprintf(stderr, "requested size unsuppored by the X server configuration\n");
			XFree(modes);
			return -1;
		}

		if(!XF86VidModeSwitchToMode(dpy, scr, modes[i])) {
			fprintf(stderr, "failed to set requested size with xf86vm\n");
			XFree(modes);
			return -1;
		}
		XF86VidModeSetViewPort(dpy, scr, 0, 0);

		XFree(modes);
		return 0;
	}
#endif

	fprintf(stderr, "not possible to change display size on this X server\n");
	return -1;
}
